# Checkpoint Compare

## 介绍
在模型训练过程中或结束后，可能保存一些检查点文件(checkpoint，简称ckpt)记录当前模型、优化器等训练状态, 工具支持比较两个不同的ckpt，评估模型相似度。

当前支持Megatron-LM、MindSpeed(PyTorch/MindTorch)的ckpt比较。支持TP、PP、EP、VPP模型并行；支持megatron.core、megatron.legacy、TransformerEngine的模型实现。


## 安装教程

参见 msprobe [安装教程](./01.installation.md)

## 使用说明
Megatron、MindSpeed的ckpt加载依赖megatron，请确保megatron在python环境中或megatron在当前路径下。


启动命令如下
```shell
msprobe --framework pytorch config_check --compare path1 path2 -o output_path.json
```

| 参数名 | 解释 | 是否必选 |
|--------|-------|--------|
| -f 或 --framework | 深度学习框架，str类型。支持参数：pytorch,mindspore，注意：msadaptor场景传入mindspore。 | 是 |
| -c 或 --compare | 2个ckpt的路径 | 是 |
| -o 或 --output | 比对结果输出路径，默认为 ./ckpt_similarity.json。输出路径存在时将报错终止。 | 否 |

Megatron-LM 和 MindSpeed 的 ckpt 目录结构如下：

```txt
directory_name/
├── iter_0000005/ # 某个iteration时的ckpt目录。
│   └── mp_rank_xx_xxx/  # 单个rank的ckpt目录，xx_xxx为模型并行索引。
│       └── model_optim_rng.pt   # 包含模型参数、随机状态等的PyTorch binary文件。
├── iter_0000010/
├── latest_checkpointed_iteration.txt  # 记录最后一个保存的ckpt的纯文本文件。
```

对于--compare参数的两个路径，为directory_name时，工具通过latest_checkpointed_iteration.txt自动选择latest checkpoint进行比对. 为directory_name/iter_xxxxxxx时, 工具使用指定iteration的ckpt进行比对。暂不支持单个rank的比对。

## 输出示例
Checkpoint比对结果以json文件输出，内容如下示例：
```json
{
    "decoder.layers.0.input_layernorm.weight": {
        "l2": 0.0, 
        "cos": 0.999999,
        "numel": 128,
        "shape": [
            128
        ]
    },
    "decoder.layers.0.pre_mlp_layernorm.weight": {
        "l2": 0.012, 
        "cos": 0.98,
        "numel": 128,
        "shape": [
            128
        ]
    }
}
```

统计量 | 解释 |
|-------|---------|
| l2 | 欧式距离，$\|\|a-b\|\|_2$ |
| cos | 余弦相似度， $\frac{<a,b>}{\|\|a\|\|_2\|\|b\|\|_2}$ |
| numel | 参数的元素个数 |
| shape | 参数的shape |