# 强化学习数据采集

## 介绍
在强化学习训练过程中，往往存在多个模型（actor、reward、reference）和两个阶段（推理、训练），问题定界困难。

本工具提供一种灵活存储强化学习训练过程中关键阶段性数据的能力，并支持对比两次采集的关键数据，以支持问题快速定界。

常用关键数据示例：prompt、response、reward、log_prob、ref_log_probe、old_log_probe、kl_loss。


## 安装教程

参见 msprobe [安装教程](./01.installation.md)。

## 使用说明

### 数据采集

用户识别脚本中需要采集数据的地方，然后通过插入代码的方式采集关键数据。

当确定需要采集数据的地方，例如response，可以按如下方式对数据进行存储：
```
from msprobe.core import SingleSave
SingleSave("./dump_path", fmk="pytorch")
SingleSave.save({"response": response})
```
其中"./dump_path"为输出路径，没有默认值，需要自己配置；fmk可选"pytorch"或者"mindspore"，默认"pytorch"。

其中"response"是可以任意指定的key，response是训练过程中的真实tensor变量。

也支持一次性存储多个数据：
```
from msprobe.core import SingleSave
SingleSave("./dump_path", fmk="pytorch")
SingleSave.save({
    "prompt": prompt,
    "response": response
    })
```

### 配置保存

当确定需要采集数据配置json的地方，可以按如下方式对配置进行存储：
```
from msprobe.core import SingleSave
SingleSave("./dump_path")
SingleSave.save_config(configurations_json)
```


采集到的数据目录结构如下：
```txt
dump_path/
├── data/ # 固定为data
│   └── response/  # 关键数据名称，来自SingleSave.save的时候的key
│           └── step0/  # step数
│               └── rank0/   # rank数
│                     └── micro_step0/   #micro_step数
|                              └── response0.npy   #存储的关键数据的真实npy文件
|                              └── response0.json  #存储的关键数据的统计量文件，包括tensor的最大、最小、均值、norm、shape
├── configurations.json  # 配置json文件
```

### 结果比对

两次采集数据之后得到dump_path1和dump_path2，可以创建一个比对脚本，例如compare.py，将两次训练的dump_path传入：
```
from msprobe.core import SingleComparator
SingleComparator.compare(
    "dump_path1", 
    "dump_path2", 
    "output_path")
```

会在output_path下对每种关键数据都生成excel结果表格，比如response.xlsx，形式为关键数据的名字加上.xlsx后缀。

表格会体现每一个对应tensor的差异，解释：

表头 | 解释 |
|-------|---------|
| step | 训练步数 |
| rank | 卡号 |
| micro_step | 梯度累计步数 |
| id | 参数的shape |
| shape1 | dump_path1中的数据形状 |
| shape2 | dump_path2中的数据形状 |
| 相同元素百分比 | 元素相同的个数占总元素个数的百分比 |
| 首个不匹配元素索引 | 首个匹配不上的元素是第几个 |
| 最大绝对误差 | 最大绝对误差 |
| 最大相对误差 | 最大相对误差 |
| 误差在千分之一内元素占比 | 误差在千分之一内元素个数占总元素个数的百分比 |
| 误差在百分之一内元素占比 | 误差在百分之一内元素个数占总元素个数的百分比 |
